import { injectable, inject } from 'inversify';
import * as lsp from 'vscode-languageserver';
import { EventTypes, TspClient, CommandTypes } from './tsp-client';
import {
    completionKindsMapping,
    toDocumentHighlight, toSymbolKind, toPosition, toLocation, toTextEdit,
} from '../utils';
import { TextDocument, uriToPath, convertChangesRange, pathToUri } from '@workbench-stack/language-core';

@injectable()
export class JavascriptService {
    private initializeParams;
    private initializeResult;
    private tspClient;
    private openedDocumentUris: Map<string, any> = new Map<string, any>();
    private diagnosticQueue;
    public readonly documents;

    public onInitialize = (params) => {
        console.log('The js/ts server is initialized.');
        this.initializeParams = params;

        this.tspClient = new TspClient({
            tsserverPath: 'tsserver',
            onEvent: event => { },
        });

        this.tspClient.start();

        return {
            capabilities: {
                completionProvider: true,
                definitionProvider: true,
                documentFormattingProvider: true,
                documentHighlightProvider: true,
                documentSymbolProvider: true,
                hoverProvider: false,
                renameProvider: true,
                referencesProvider: true,

                workspaceSymbolProvider: false,
                documentLinkProvider: false,
            },
        };
    }

    public onDidOpenTextDocument = async params => {
        console.log('onDidOpenTextDocument js/ts');
        const path = uriToPath(params.textDocument.uri);
        const document = new TextDocument(params.textDocument);
        this.openedDocumentUris.set(params.textDocument.uri, document);
        this.tspClient.notify(CommandTypes.Open, {
            file: path,
            fileContent: params.textDocument.text,
        });
    }

    public onDidChangeTextDocument = params => {
        console.log('onDidChangeTextDocument js/ts');
        const document = this.getOpenDocument(params.textDocument.uri);
        const changes = convertChangesRange(params.contentChanges);

        document.apply(changes, params.textDocument.version);
    }

    public onDidCloseTextDocument = async (params) => {
        console.log('onDidCloseTextDocument js/ts');
        const path = uriToPath(params.textDocument.uri);
        this.openedDocumentUris.delete(params.textDocument.uri);
        if (this.tspClient) {
           this.tspClient.notify(CommandTypes.Close, { file: path });
        }
    }

    public async onRename(params) {
        console.log('onRenameRequest js/ts', params);
        const path = uriToPath(params.textDocument.uri);
        const result = await this.tspClient.request(CommandTypes.Rename, {
            file: path,
            line: params.position.line + 1,
            offset: params.position.character + 1,
        });
        const workspaceEdit = {
            changes: {},
        };

        if (result.body.info.localizedErrorMessage) {
            return {
                rejectReason: result.body.info.localizedErrorMessage,
                edits: result.body.locs,
            };
        }
        if (!result.body || !result.body.info.canRename || result.body.locs.length === 0) {
            return workspaceEdit;
        }
        result.body.locs
            .forEach((spanGroup) => {
                const uri = pathToUri(spanGroup.file),
                    textEdits = workspaceEdit.changes[uri] || (workspaceEdit.changes[uri] = []);

                spanGroup.locs.forEach((textSpan) => {
                    textEdits.push({
                        newText: params.newName,
                        range: {
                            start: toPosition(textSpan.start),
                            end: toPosition(textSpan.end),
                        },
                    });
                });
            });
        return workspaceEdit;
    }

    public async onReferences(params) {
        console.log('onReferences js/ts');
        const path = uriToPath(params.textDocument.uri);
        try {
            const result = await this.tspClient.request(CommandTypes.References, {
                file: path,
                line: params.position.line + 1,
                offset: params.position.character + 1,
            });

            if (!result.body) {
                return [];
            }
            return result.body.refs
                .map(fileSpan => toLocation(fileSpan));
        } catch (err) {
            console.log('err', err);
            return [];
        }
    }

    public async onHover(params) {
        console.log('onHover js/ts');
        const path = uriToPath(params.textDocument.uri);

        let result;
        try {
            result = await this.tspClient.request(CommandTypes.Quickinfo, {
                file: path,
                line: params.position.line + 1,
                offset: params.position.character + 1,
            });

            if (!result.body) {
                return <lsp.Hover>{
                    contents: [],
                };
            }

            const range = {
                start: toPosition(result.body.start),
                end: toPosition(result.body.end),
            };

            const contents: lsp.MarkedString[] = [
                { language: 'typescript', value: result.body.displayString },
            ];
            if (result.body.documentation) {
                contents.push(result.body.documentation);
            }

            return {
                contents,
                range,
            };
        } catch (err) {
            console.log('err', err);
            return <lsp.Hover>{
                contents: [],
            };
        }
    }
    public async onDocumentSymbol(params) {
        console.log('onDocumentSymbol js/ts');
        const path = uriToPath(params.textDocument.uri);

        const response = await this.tspClient.request(CommandTypes.NavTree, {
            file: path,
        });

        if (!response.body) {
            return [];
        }
        const result = [];
        const collectSymbol: (element, parent: string | undefined, acceptor: (sy) => void) => void =
            (element, parent, acceptor) => {
                const start = element.spans[0];
                const end = element.spans[element.spans.length - 1];
                if (start && end) {
                    const symbol = lsp.SymbolInformation.create(
                        element.text,
                        toSymbolKind(element.kind),
                        { start: toPosition(start.start), end: toPosition(end.end) },
                        params.textDocument.uri,
                        parent,
                    );
                    acceptor(symbol);
                }
                if (element.childItems) {
                    for (const child of element.childItems) {
                        collectSymbol(child, element.text, acceptor);
                    }
                }
            };
        collectSymbol(response.body, undefined, sym => result.push(sym));

        return result.map(item => {
            if (!item.containerName) {
                item.containerName = '';
            }
            return item;
        });
    }
    public async onDocumentHighlight(arg) {
        console.log('onDocumentHighlight js/ts');
        let response;
        try {
            response = await this.tspClient.request(CommandTypes.DocumentHighlights, {
                file: uriToPath(arg.textDocument.uri),
                line: arg.position.line + 1,
                offset: arg.position.character + 1,
                filesToSearch: [uriToPath(arg.textDocument.uri)],
            });
        } catch (err) {
            return [];
        }
        if (!response.body) {
            return [];
        }
        const result = [];
        for (const item of response.body) {
            const highlights = toDocumentHighlight(item);
            result.push(...highlights);
        }
        return result;
    }

    public async onDocumentFormatting(params) {
        console.log('onDocumentFormatting js/ts');
        const path = uriToPath(params.textDocument.uri);

        let opts = {
            ...params.options,
        };
        // translate
        if (!opts.convertTabsToSpaces) {
            opts.convertTabsToSpaces = params.options.insertSpaces;
        }

        const response = await this.tspClient.request(CommandTypes.Format, {
            file: path,
            line: 1,
            offset: 1,
            endLine: Number.MAX_SAFE_INTEGER,
            endOffset: Number.MAX_SAFE_INTEGER,
            options: {
                'tabSize': 4,
                'indentSize': 4,
                'newLineCharacter': '\n',
                'convertTabsToSpaces': false,
                'insertSpaceAfterCommaDelimiter': true,
                'insertSpaceAfterSemicolonInForStatements': true,
                'insertSpaceBeforeAndAfterBinaryOperators': true,
                'insertSpaceAfterKeywordsInControlFlowStatements': true,
                'insertSpaceAfterFunctionKeywordForAnonymousFunctions': true,
                'insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis': false,
                'insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets': false,
                'insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces': false,
                'insertSpaceBeforeFunctionParenthesis': false,
                'placeOpenBraceOnNewLineForFunctions': false,
                'placeOpenBraceOnNewLineForControlBlocks': false,
            },
        });

        if (response.body) {
            return response.body.map(e => toTextEdit(e));
        }
        return [];
    }

    public async onDefinition(params) {
        console.log('onDefinition js/ts');
        const path = uriToPath(params.textDocument.uri);

        const result = await this.tspClient.request(CommandTypes.Definition, {
            file: path,
            line: params.position.line + 1,
            offset: params.position.character + 1,
        });

        return result.body ? result.body
            .map(fileSpan => toLocation(fileSpan)) : [];
    }
    public async onCompletionResolve(item) {
        console.log('onCompletionResolve js/ts');
        const result = await this.tspClient.request(CommandTypes.CompletionDetails, {
            entryNames: [item.label],
            file: item.data.file as string,
            line: item.data.line as number,
            offset: item.data.offset as number,
        });
        if (!result.body) {
            return item;
        }
        if (result.body[0] && result.body[0].documentation) {
            item.documentation = result.body[0].documentation.map(i => i.text).join('\n');
        }
        return item;
    }

    public async onCompletion(params) {
        console.log('onCompletion js/ts');
        const path = uriToPath(params.textDocument.uri);
        try {
            const result = await this.tspClient.request(CommandTypes.Completions, {
                file: path,
                line: params.position.line + 1,
                offset: params.position.character + 1,
                prefix: '',
            });

            return {
                isIncomplete: false,
                items: result.body ? result.body
                    .map(item => {
                        return {
                            label: item.name,
                            kind: completionKindsMapping[item.kind],
                            // store information for resolve
                            data: {
                                file: path,
                                line: params.position.line + 1,
                                offset: params.position.character + 1,
                            },
                        };
                    }) : [],
            };
        } catch (error) {
            console.log('error', error);
            return [];
        }
    }

    private getOpenDocument(uri: string): TextDocument {
        const doc = this.openedDocumentUris.get(uri);
        if (doc) {
            return doc;
        } else {
            throw new Error('Document ' + uri + ' has not been opened.');
        }
    }
}
